#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "lib/game_logic.h"
#include "lib/config.h"

bool game_state_init(game_state* state, config* conf) {
  srand(time(NULL));
  state->width = conf->width;
  state->height = conf->height;
  state->mines = conf->mines;
  state->flags = conf->flags;
  state->status = RUNNING;

  const uint32_t map_size = state->width * state->height;
  state->map = (field*)malloc(map_size * sizeof(field));
  if (!state->map) {
    fprintf(stderr, "Impossible to allocate memory for the map");
    return false;
  }

  for (uint32_t i = 0; i < map_size; ++i) {
    state->map[i].state = CLOSED;
    state->map[i].is_protected = false;
    state->map[i].near_bombs = 0;
  }

  if (state->mines > map_size) {
    fprintf(stderr, "Too many mines for this field size");
    free(state->map);
    return false;
  }
  uint32_t mines_placed = 0;
  while (mines_placed < state->mines) {
    uint32_t index = rand() % map_size;
    if (state->map[index].state != MINED) {
      state->map[index].state = MINED;
      mines_placed++;
    }
  }
  return true;
}

bool game_state_free(game_state* state) {
  if (!state) {
    fprintf(stderr, "Impossible to clear the game state. The pointer is NULL");
    return false;
  }
  if (!state->map) {
    fprintf(stderr, "Impossible to clear the map. The pointer is NULL");
    return false;
  }
  free(state->map);
  state->map = NULL;


  state->width = 0;
  state->height = 0;
  state->mines = 0;
  return true;
}

void game_state_debug_print(const game_state* state) {
  printf("Width: %u\n", state->width);
  printf("Height: %u\n", state->height);
  printf("Mines: %u\n", state->mines);
  printf("Flags: %u\n", state->flags);

  for (uint32_t y = 0; y < state->height; ++y) {
    for (uint32_t x = 0; x < state->width; ++x) {
      const uint32_t index = y * state->width + x;
      printf("%c ", state->map[index].state);
    }
    printf("\n");
  }
  printf("\n");

  for (uint32_t y = 0; y < state->height; ++y) {
    for (uint32_t x = 0; x < state->width; ++x) {
      printf("%d ", state->map[y * state->width + x].near_bombs);
    }
    printf("\n");
  }
  printf("\n");
}

void calculate_near_bombs(game_state* state) {
  uint32_t width = state->width;
  uint32_t height = state->height;
    
  for (uint32_t y = 0; y < height; ++y) {
    for (uint32_t x = 0; x < width; ++x) {
      int32_t i = y * width + x;
      uint8_t bombs = 0;
      if (state->map[i].state == MINED) continue;
            
      for (int32_t dy = -1; dy <= 1; ++dy) {
        for (int32_t dx = -1; dx <= 1; ++dx) {
          if (dx == 0 && dy == 0) continue;
                    
          int32_t nx = x + dx;
          int32_t ny = y + dy;
                    
          if (nx >= 0 && nx < width && ny >= 0 && ny < height) {
            int32_t neighbor_idx = ny * width + nx;
            if (state->map[neighbor_idx].state == MINED) {
              bombs++;
            }
          }
        }
      }
      state->map[i].near_bombs = bombs;
    }
  }
}

void open_field(game_state* state, uint32_t x, uint32_t y) {
  if (x >= state->width || y >= state->height) {
    return;
  }

  const uint32_t index = y * state->width + x;
  field* current = &state->map[index];

  if (current->is_protected || current->state == OPENED) {
    return;
  }

  if (current->state == MINED) {
    current->state = DESTROYED;
    state->status = LOSE;
    return;
  }

  current->state = OPENED;

  if (current->near_bombs == 0) {
    for (int dy = -1; dy <= 1; ++dy) {
      for (int dx = -1; dx <= 1; ++dx) {
        if (dx == 0 && dy == 0) continue;

        const int32_t nx = x + dx;
        const int32_t ny = y + dy;

        if (nx >= 0 && nx < state->width &&
            ny >= 0 && ny < state->height) {
          open_field(state, nx, ny);
        }
      }
    }
  }
}

void set_flag(game_state* state, uint32_t x, uint32_t y) {
  if (x >= state->width || y >= state->height) {
    return;
  }

  const uint32_t index = y * state->width + x;
  field* current = &state->map[index];

  if (current->is_protected) {
    current->is_protected = false;
    state->flags++;
    return;
  }

  switch (current->state) {
  case MINED:
  case CLOSED:
    if (state->flags > 0) {
      current->is_protected = true;
      state->flags--;
    }
    break;
  case OPENED:
  case DESTROYED:
  default:
    return;
  }
}

void check_win(game_state* state) {
  if (state->status == LOSE) return;
  uint32_t safe_fields_opened = 0;
  uint32_t mines_flagged = 0;

  const uint32_t map_size = state->width * state->height;

  for (uint32_t i = 0; i < map_size; ++i) {
    const field* field = &state->map[i];

    if (field->state == MINED) {
      if (field->is_protected) {
        mines_flagged++;
      }
    } else if (field->state == OPENED) {
      safe_fields_opened++;
    }
  }

  const uint32_t total_safe_fields = map_size - state->mines;

  if ((safe_fields_opened == total_safe_fields)
      && (mines_flagged == state->mines)) {
    state->status = WIN;
  }
}

void make_action(game_state* state, uint32_t x, uint32_t y, game_operations op) {
  switch (op) {
  case OPEN:
    open_field(state, x, y);
    break;
  case SET_FLAG:
    set_flag(state, x, y);
    break;
  default:
    fprintf(stderr, "Unknown operation: %c", op);
  }
  check_win(state);
}
